pub fn select_from_list<F: FnOnce(&str) -> String>(
    prompt: &str,
    lst: Vec<&str>,
    quit: bool,
    then: F,
) -> String {
    match inquire::Select::new(prompt, lst).prompt() {
        Ok(val) => then(val),
        Err(_) => {
            if quit {
                eprintln!("Nothing selected");
                std::process::exit(1);
            }

            String::new()
        }
    }
}

pub fn select_gitmoji() -> String {
    let topics = vec![
        "Fix",
        "Refactor",
        "Improve",
        "Add/Update",
        "Remove",
        "Git",
        "UI",
        "Dependency",
        "Security",
        "Text",
        "Test",
        "Fun",
        "Misc",
    ];

    let select = select_from_list(
        "Gitmoji Intention Topic",
        topics,
        true,
        |topic| match topic {
            "Fix" => select_gitmoji_fix(),
            "Improve" => select_gitmoji_improvement(),
            "Remove" => select_gitmoji_remove(),
            "Dependency" => select_gitmoji_dependency(),
            "Test" => select_gitmoji_test(),
            "Git" => select_gitmoji_git(),
            "Refactor" => select_gitmoji_refactor(),
            "UI" => select_gitmoji_ui(),
            "Text" => select_gitmoji_text(),
            "Fun" => select_gitmoji_fun(),
            "Security" => select_gitmoji_security(),
            "Add/Update" => select_gitmoji_add_update(),
            _ => select_gitmoji_misc(),
        },
    );

    if select.is_empty() {
        return select_gitmoji();
    }

    select
}

pub fn select_gitmoji_fix() -> String {
    let options = vec![
        "🐛 Fix a bug",
        "🩹 Simple fix for a non-critical issue",
        "🚑️ Critical hotfix",
        "✏️ Fix typos",
        "💚 Fix CI Build",
        "🔒️ Fix security or privacy issues",
        "🚨 Fix compiler / linter warnings",
        "🥅 Catch errors",
    ];

    let res = select_from_list("Gitmoji Intention", options, false, |selected| {
        selected.split_whitespace().next().unwrap().to_string()
    });

    if res.is_empty() {
        return select_gitmoji();
    }

    res
}

pub fn select_gitmoji_improvement() -> String {
    let options = vec![
        "✨ Introduce new features",
        "🎨 Improve structure / format of the code",
        "⚡️ Improve performance",
        "🧑‍💻 Improve developer experience",
        "🚸 Improve user experience / usability",
        "♿️ Improve accessibility",
        "🔍️ Improve SEO",
    ];

    let res = select_from_list("Gitmoji Intention", options, false, |selected| {
        selected.split_whitespace().next().unwrap().to_string()
    });

    if res.is_empty() {
        return select_gitmoji();
    }

    res
}

pub fn select_gitmoji_remove() -> String {
    let options = vec!["🔥 Remove code or files", "⚰️ Remove dead code"];

    let res = select_from_list("Gitmoji Intention", options, false, |selected| {
        selected.split_whitespace().next().unwrap().to_string()
    });

    if res.is_empty() {
        return select_gitmoji();
    }

    res
}

pub fn select_gitmoji_dependency() -> String {
    let options = vec![
        "➖ Remove a dependency",
        "➕ Add a dependency",
        "⬇️ Downgrade dependencies",
        "⬆️ Upgrade dependencies",
        "📌 Pin dependencies to specific versions",
    ];

    let res = select_from_list("Gitmoji Intention", options, false, |selected| {
        selected.split_whitespace().next().unwrap().to_string()
    });

    if res.is_empty() {
        return select_gitmoji();
    }

    res
}

pub fn select_gitmoji_test() -> String {
    let options = vec![
        "✅ Add, update, or pass tests",
        "🧪 Add a failing test",
        "🤡 Mock things",
        "⚗️ Perform experiments",
    ];

    let res = select_from_list("Gitmoji Intention", options, false, |selected| {
        selected.split_whitespace().next().unwrap().to_string()
    });

    if res.is_empty() {
        return select_gitmoji();
    }

    res
}

pub fn select_gitmoji_fun() -> String {
    let options = vec![
        "🍻 Write code drunkenly",
        "🍃 Write code high",
        "❄ Write code on coke",
        "🍄 Write code on psychedelics",
        "🥚 Add or update an easter egg",
        "💩 Write bad code that needs to be improved",
    ];

    let res = select_from_list("Gitmoji Intention", options, false, |selected| {
        selected.split_whitespace().next().unwrap().to_string()
    });

    if res.is_empty() {
        return select_gitmoji();
    }

    res
}

pub fn select_gitmoji_git() -> String {
    let options = vec![
        "⏪️ Revert changes",
        "🔀 Merge branches",
        "🔖 Release / Version tags",
        "🙈 Add or update a .gitignore file",
    ];

    let res = select_from_list("Gitmoji Intention", options, false, |selected| {
        selected.split_whitespace().next().unwrap().to_string()
    });

    if res.is_empty() {
        return select_gitmoji();
    }

    res
}

pub fn select_gitmoji_refactor() -> String {
    let options = vec![
        "♻️ Refactor code",
        "🗑️ Deprecate code that needs to be cleaned up",
        "🚚 Move or rename resources (e.g.: files, paths, routes)",
        "🍱 Add or update assets",
        "👽️ Update code due to external API changes",
        "💡 Add or update comments in source code",
        "🏷️ Add or update types",
    ];

    let res = select_from_list("Gitmoji Intention", options, false, |selected| {
        selected.split_whitespace().next().unwrap().to_string()
    });

    if res.is_empty() {
        return select_gitmoji();
    }

    res
}

pub fn select_gitmoji_ui() -> String {
    let options = vec![
        "💄 Add or update the UI and style files",
        "📱 Work on responsive design",
        "💫 Add or update animations and transitions",
    ];

    let res = select_from_list("Gitmoji Intention", options, false, |selected| {
        selected.split_whitespace().next().unwrap().to_string()
    });

    if res.is_empty() {
        return select_gitmoji();
    }

    res
}

pub fn select_gitmoji_text() -> String {
    let options = vec![
        "📝 Add or update documentation",
        "🌐 Internationalization and localization",
        "💬 Add or update text and literals",
        "👥 Add or update contributor(s)",
        "📄 Add or update license",
        "🔊 Add or update logs",
        "🔇 Remove logs",
    ];

    let res = select_from_list("Gitmoji Intention", options, false, |selected| {
        selected.split_whitespace().next().unwrap().to_string()
    });

    if res.is_empty() {
        return select_gitmoji();
    }

    res
}

pub fn select_gitmoji_security() -> String {
    let options = vec![
        "🛂 Work on code related to authorization, roles and permissions",
        "🔐 Add or update secrets",
        "🦺 Add or update code related to validation",
    ];

    let res = select_from_list("Gitmoji Intention", options, false, |selected| {
        selected.split_whitespace().next().unwrap().to_string()
    });

    if res.is_empty() {
        return select_gitmoji();
    }

    res
}

pub fn select_gitmoji_add_update() -> String {
    let options = vec![
        "👷 Add or update CI build system",
        "🔧 Add or update configuration files",
        "🔨 Add or update development scripts",
        "📦️ Add or update compiled files or packages",
        "🚩 Add, update, or remove feature flags",
        "🩺 Add or update healthcheck",
        "🧵 Add or update code related to multithreading or concurrency",
        "🌱 Add or update seed files",
        "📸 Add or update snapshots",
        "📈 Add or update analytics or track code",
        "👔 Add or update business logic",
    ];

    let res = select_from_list("Gitmoji Intention", options, false, |selected| {
        selected.split_whitespace().next().unwrap().to_string()
    });

    if res.is_empty() {
        return select_gitmoji();
    }

    res
}

pub fn select_gitmoji_misc() -> String {
    let options = vec![
        "🎉 Begin a project",
        "🚀 Deploy stuff",
        "🚧 Work in progress",
        "💥 Introduce breaking changes",
        "🗃️ Perform database related changes",
        "🏗️ Make architectural changes",
        "🧱 Infrastructure related changes",
        "💸 Add sponsorships or money related infrastructure",
        "🧐 Data exploration/inspection",
    ];

    let res = select_from_list("Gitmoji Intention", options, false, |selected| {
        selected.split_whitespace().next().unwrap().to_string()
    });

    if res.is_empty() {
        return select_gitmoji();
    }

    res
}

pub struct CommitMessage {
    gitmoji: String,
    title: String,
    body: Option<String>,
}

impl CommitMessage {
    pub const fn new(gitmoji: String, title: String, body: Option<String>) -> Self {
        CommitMessage {
            gitmoji,
            title,
            body,
        }
    }

    pub fn to_msg(&self) -> String {
        let mut msg = self.gitmoji.clone();
        msg.push_str(&format!(" {}", self.title));

        if let Some(body) = &self.body {
            if !body.is_empty() {
                msg.push_str(&format!("\n\n{body}"));
            }
        }

        msg
    }
}
